#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "st_uvc.h"
#include "myuvc.h"

unsigned int g_MaxFrameLength =0;
FILE *g_ImageEs_fp = NULL;
FILE *g_ImageSz_fp = NULL;

#ifdef UVC_SUPPORT_LOWLATENCY
#define Piece_TransferSize 2048
#else
#define Piece_TransferSize 0
#endif

typedef struct usr_buf_s {
    void *userBuf;
    unsigned int bufsize;
    unsigned int length;
    unsigned int bytesused;
    double framerate;
} usr_buf_t;

usr_buf_t userBuf;

const char IMG_PATH[] ="../src/sample_dummy/img";

static unsigned int get_sys_time()
{
    struct timespec t1; 
    clock_gettime(CLOCK_MONOTONIC, &t1);
    return 1000000 * (t1.tv_sec) + (t1.tv_nsec) / 1000;
}

static void frameRate_control(double framerate)
{
    static unsigned int old = 0;
    unsigned int interval = 1.0 / framerate * 1000000; 

    while( interval >get_sys_time() - old){
        usleep(100);
    }
    old = get_sys_time();
}

static inline char *get_file_name(char *filename,unsigned int width,unsigned int height)
{
    char str4num[10];
    
    if(!filename)
        return NULL;

    memset(filename,0x00,strlen(filename));
    memset(str4num,0x00,strlen(str4num));
    sprintf(str4num,"%d",width);
    strcat(filename,str4num);
    strcat(filename,"_");
    memset(str4num,0x00,strlen(str4num));
    sprintf(str4num,"%d",height);
    strcat(filename,str4num);

    return filename;
}

static void destroy_filp()
{
    if(g_ImageEs_fp){
       fclose(g_ImageEs_fp);
       g_ImageEs_fp = NULL;
    }

    if(g_ImageSz_fp){
       fclose(g_ImageSz_fp);
       g_ImageSz_fp = NULL;
    }
}

static int fill_filp_with_format(unsigned int fcc,unsigned int width,unsigned int height)
{
    char EsFileName[80],SzFileName[80];
    char EsFileEnd[20],SzFileEnd[20];
    char FileDir[80],FileName[60];

    memset(EsFileName,0x00,strlen(EsFileName));
    memset(SzFileName,0x00,strlen(SzFileName));
    memset(EsFileEnd,0x00,strlen(EsFileEnd));
    memset(SzFileEnd,0x00,strlen(SzFileEnd));
    memset(FileDir,0x00,strlen(FileDir));
    memset(FileName,0x00,strlen(FileName));

    strcat(FileDir,IMG_PATH);

    if(!get_file_name(FileName,width,height)){
        printf("%s get_file_name failed\n",__func__);
        return -EINVAL;
    }

    switch(fcc){
        case V4L2_PIX_FMT_H264:
            strcat(FileDir,"/H264/");
            strcat(EsFileEnd,".es");
            strcat(SzFileEnd,".size");
            break;
        case V4L2_PIX_FMT_H265:
            strcat(FileDir,"/H265/");
            strcat(EsFileEnd,".es");
            strcat(SzFileEnd,".size");
            break;
        case V4L2_PIX_FMT_MJPEG:
            strcat(FileDir,"/MJPEG/");
            //strcat(EsFileEnd,".pic");
            //strcat(SzFileEnd,".picsize");
            strcat(EsFileEnd,".es");
            strcat(SzFileEnd,".size");
            break;
        case V4L2_PIX_FMT_NV12:
            strcat(FileDir,"/NV12/");
            strcat(EsFileEnd,".es");
            strcat(SzFileEnd,".size");
            break;
        case V4L2_PIX_FMT_YUYV:
            strcat(FileDir,"/YUYV/");
            strcat(EsFileEnd,".es");
            strcat(SzFileEnd,".size");
            break;
        default:
            return -EINVAL;
    }

    strcat(EsFileName,FileDir);
    strcat(SzFileName,FileDir);
    strcat(EsFileName,FileName);
    strcat(SzFileName,FileName);
    strcat(EsFileName,EsFileEnd);
    strcat(SzFileName,SzFileEnd);
    printf("Es Path: %s \nSz Path: %s \n",EsFileName,SzFileName);

    g_ImageEs_fp = fopen(EsFileName,"rb");
    g_ImageSz_fp = fopen(SzFileName,"rb");
    if(!g_ImageEs_fp || !g_ImageSz_fp){
        printf("Please run at the bin dir ERR: %d\n",errno);
    }

    return 0;
}

int UVC_DEMO_Init(void *uvc)
{
    if(NULL==uvc){
        printf(" %s Invalid Argument \n",__func__);
        return -1;
    }

    return 0;
}

int UVC_DEMO_Deinit(void *uvc)
{
    if(NULL==uvc){
        printf(" %s Invalid Argument \n",__func__);
        return -1;
    }
    ST_UVC_Device_t *uvc_dev = (ST_UVC_Device_t*)uvc;


    if(uvc_dev->Input_Device){ 
        uvc_dev->Input_Device = NULL;
    }
    return 0;
}

int UVC_DEMO_UP_FillBuffer(void *uvc,ST_UVC_BufInfo_t *bufInfo)
{
    unsigned int Szlen = 0,Eslen;

    if(!g_ImageEs_fp || !g_ImageSz_fp)
        goto err;
retry:
    if(userBuf.bytesused >= userBuf.length){ // Get A New Frame
        memset(userBuf.userBuf, 0x00, userBuf.bufsize);
        if(4!=fread(&Szlen,1,4,g_ImageSz_fp)){
            fseek(g_ImageSz_fp, 0, SEEK_SET);
            fseek(g_ImageEs_fp, 0, SEEK_SET);
            goto retry;
        }
        Eslen = fread(userBuf.userBuf, 1, Szlen, g_ImageEs_fp);
        if(Szlen!=Eslen){
            fseek(g_ImageSz_fp, 0, SEEK_SET);
            fseek(g_ImageEs_fp, 0, SEEK_SET);
            goto retry;
        }
        userBuf.length = Eslen;
        userBuf.bytesused = 0;
    }
    if(Piece_TransferSize){ // Lowlatency Mode
        if(userBuf.bytesused + Piece_TransferSize >= userBuf.length) {
            memcpy(bufInfo->b.buf, userBuf.userBuf + userBuf.bytesused, userBuf.length - userBuf.bytesused);
            bufInfo->is_tail = true;
            bufInfo->length = userBuf.length - userBuf.bytesused;
            userBuf.bytesused += userBuf.length - userBuf.bytesused;
        }
        else{
            memcpy(bufInfo->b.buf, userBuf.userBuf + userBuf.bytesused, Piece_TransferSize);
            bufInfo->is_tail = false;
            bufInfo->length = Piece_TransferSize;
            userBuf.bytesused += Piece_TransferSize;
        }
    }
    else{
        memcpy(bufInfo->b.buf,userBuf.userBuf,userBuf.length);
        bufInfo->length = userBuf.length;
        userBuf.bytesused += userBuf.length;
    }
    frameRate_control(userBuf.framerate);
    return bufInfo->length;
err:
    return -EINVAL;
}

int UVC_DEMO_MM_FillBuffer(void *uvc, ST_UVC_BufInfo_t *bufInfo)
{
    unsigned int Szlen = 0,Eslen;

    if(!g_ImageEs_fp || !g_ImageSz_fp)
        goto err;
retry:
    if(userBuf.bytesused >= userBuf.length){ // Get A New Frame
        memset(userBuf.userBuf, 0x00, userBuf.bufsize);
        if(4!=fread(&Szlen,1,4,g_ImageSz_fp)){
            fseek(g_ImageSz_fp, 0, SEEK_SET);
            fseek(g_ImageEs_fp, 0, SEEK_SET);
            goto retry;
        }
        Eslen = fread(userBuf.userBuf, 1, Szlen, g_ImageEs_fp);
        if(Szlen!=Eslen){
            fseek(g_ImageSz_fp, 0, SEEK_SET);
            fseek(g_ImageEs_fp, 0, SEEK_SET);
            goto retry;
        }
        userBuf.length = Eslen;
        userBuf.bytesused = 0;
    }
    if(Piece_TransferSize){ // Lowlatency Mode
        if(userBuf.bytesused + Piece_TransferSize >= userBuf.length) {
            memcpy(bufInfo->b.buf, userBuf.userBuf + userBuf.bytesused, userBuf.length - userBuf.bytesused);
            bufInfo->is_tail = true;
            bufInfo->length = userBuf.length - userBuf.bytesused;
            userBuf.bytesused += userBuf.length - userBuf.bytesused;
        }
        else{
            memcpy(bufInfo->b.buf, userBuf.userBuf + userBuf.bytesused, Piece_TransferSize);
            bufInfo->is_tail = false;
            bufInfo->length = Piece_TransferSize;
            userBuf.bytesused += Piece_TransferSize;
        }
    }
    else{
        memcpy(bufInfo->b.buf,userBuf.userBuf,userBuf.length);
        bufInfo->length = userBuf.length;
        userBuf.bytesused += userBuf.length;
    }
    frameRate_control(userBuf.framerate);
    return bufInfo->length;
err:
    return -EINVAL;
}

int UVC_DEMO_StartCapture(void *uvc,Stream_Params_t format)
{
    if(NULL==uvc){
        printf(" %s Invalid Argument \n",__func__);
        return -1;
    }
    
    Stream_Params_t stream_param = ((ST_UVC_Device_t*)uvc)->stream_param;
    g_MaxFrameLength = stream_param.maxframesize;

    fill_filp_with_format(format.fcc,
                          format.width,
                          format.height);

    memset(&userBuf,0x00,sizeof(userBuf));
    userBuf.userBuf = malloc(g_MaxFrameLength);
    userBuf.bufsize = g_MaxFrameLength;
    userBuf.framerate = format.frameRate;
    return 0;
}

int UVC_DEMO_StopCapture(void *uvc)
{
    if(NULL==uvc){
        printf(" %s Invalid Argument \n",__func__);
        return -1;
    }

    destroy_filp();

    if(userBuf.userBuf)
        free(userBuf.userBuf);

    return 0;
}
